@@ -348,6 +348,14 @@ buf_ValidateBuffers(void)
     }
 
     for (bp = cm_data.buf_freeListEndp; bp; bp=(cm_buf_t *) osi_QPrev(&bp->q)) {
+
+	if ( bp < (cm_buf_t *)cm_data.bufHeaderBaseAddress ||
+	     bp >= (cm_buf_t *)cm_data.bufDataBaseAddress) {
+	    afsi_log("cm_ValidateBuffers failure: out of range cm_buf_t pointers");
+	    fprintf(stderr, "cm_ValidateBuffers failure: out of range cm_buf_t pointers\n");
+	    return -11;
+	}
+
         if (bp->magic != CM_BUF_MAGIC) {
             afsi_log("cm_ValidateBuffers failure: bp->magic != CM_BUF_MAGIC");
             fprintf(stderr, "cm_ValidateBuffers failure: bp->magic != CM_BUF_MAGIC\n");
@@ -364,6 +372,14 @@ buf_ValidateBuffers(void)
     }
 
     for (bp = cm_data.buf_freeListp; bp; bp=(cm_buf_t *) osi_QNext(&bp->q)) {
+
+	if ( bp < (cm_buf_t *)cm_data.bufHeaderBaseAddress ||
+	     bp >= (cm_buf_t *)cm_data.bufDataBaseAddress) {
+	    afsi_log("cm_ValidateBuffers failure: out of range cm_buf_t pointers");
+	    fprintf(stderr, "cm_ValidateBuffers failure: out of range cm_buf_t pointers\n");
+	    return -12;
+	}
+
         if (bp->magic != CM_BUF_MAGIC) {
             afsi_log("cm_ValidateBuffers failure: bp->magic != CM_BUF_MAGIC");
             fprintf(stderr, "cm_ValidateBuffers failure: bp->magic != CM_BUF_MAGIC\n");
@@ -380,6 +396,14 @@ buf_ValidateBuffers(void)
     }
 
     for ( bp = cm_data.buf_redirListp; bp; bp = (cm_buf_t *) osi_QNext(&bp->q)) {
+
+	if ( bp < (cm_buf_t *)cm_data.bufHeaderBaseAddress ||
+	     bp >= (cm_buf_t *)cm_data.bufDataBaseAddress) {
+	    afsi_log("cm_ValidateBuffers failure: out of range cm_buf_t pointers");
+	    fprintf(stderr, "cm_ValidateBuffers failure: out of range cm_buf_t pointers\n");
+	    return -13;
+	}
+
         if (!(bp->qFlags & CM_BUF_QREDIR)) {
             afsi_log("CM_BUF_QREDIR not set on cm_buf_t in buf_redirListp");
             fprintf(stderr, "CM_BUF_QREDIR not set on cm_buf_t in buf_redirListp");
@@ -394,11 +418,27 @@ buf_ValidateBuffers(void)
     }
 
     for (bp = cm_data.buf_allp; bp; bp=bp->allp) {
+
+	if ( bp < (cm_buf_t *)cm_data.bufHeaderBaseAddress ||
+	     bp >= (cm_buf_t *)cm_data.bufDataBaseAddress) {
+	    afsi_log("cm_ValidateBuffers failure: out of range cm_buf_t pointers");
+	    fprintf(stderr, "cm_ValidateBuffers failure: out of range cm_buf_t pointers\n");
+	    return -14;
+	}
+
         if (bp->magic != CM_BUF_MAGIC) {
             afsi_log("cm_ValidateBuffers failure: bp->magic != CM_BUF_MAGIC");
             fprintf(stderr, "cm_ValidateBuffers failure: bp->magic != CM_BUF_MAGIC\n");
             return -3;
         }
+
+	if ( bp->datap < cm_data.bufDataBaseAddress ||
+	     bp->datap >= cm_data.bufEndOfData) {
+	    afsi_log("cm_ValidateBuffers failure: out of range data pointers");
+	    fprintf(stderr, "cm_ValidateBuffers failure: out of range data pointers\n");
+	    return -15;
+	}
+
         counta++;
         bpa = bp;
 
@@ -819,13 +859,17 @@ afs_uint32 buf_CleanLocked(cm_scache_t *scp, cm_buf_t *bp, cm_req_t *reqp,
     {
         release_scp = 1;
 
-        lock_ObtainWrite(&scp->rw);
-        code = cm_SyncOp(scp, NULL, bp->userp ? bp->userp : cm_rootUserp, reqp, 0,
-                         CM_SCACHESYNC_NEEDCALLBACK | CM_SCACHESYNC_GETSTATUS);
-        if (code == 0) {
-            cm_SyncOpDone(scp, NULL, CM_SCACHESYNC_NEEDCALLBACK | CM_SCACHESYNC_GETSTATUS);
+	if (scp->flags & CM_SCACHEFLAG_DELETED) {
+	    code = CM_ERROR_NOSUCHFILE;
+	} else {
+	    lock_ObtainWrite(&scp->rw);
+	    code = cm_SyncOp(scp, NULL, bp->userp ? bp->userp : cm_rootUserp, reqp, 0,
+			    CM_SCACHESYNC_NEEDCALLBACK | CM_SCACHESYNC_GETSTATUS);
+	    if (code == 0) {
+		cm_SyncOpDone(scp, NULL, CM_SCACHESYNC_NEEDCALLBACK | CM_SCACHESYNC_GETSTATUS);
+	    }
+	    lock_ReleaseWrite(&scp->rw);
         }
-        lock_ReleaseWrite(&scp->rw);
     }
 
     if (scp && (scp->flags & CM_SCACHEFLAG_DELETED)) {
@@ -2252,14 +2296,16 @@ long buf_CleanVnode(struct cm_scache *scp, cm_user_t *userp, cm_req_t *reqp)
              * it back in order to determine whether or not it is in
              * fact dirty.
              */
-            lock_ObtainRead(&buf_globalLock);
             if (bp->qFlags & CM_BUF_QREDIR) {
-                osi_Log1(buf_logp,"buf_CleanVnode buffer held by redirector bp 0x%p", bp);
+                lock_ObtainRead(&buf_globalLock);
+                if (bp->qFlags & CM_BUF_QREDIR) {
+                    osi_Log1(buf_logp,"buf_CleanVnode buffer held by redirector bp 0x%p", bp);
 
-                /* Retrieve single extent from the redirector */
-                buf_RDRShakeAnExtentFree(bp, reqp);
+                    /* Retrieve single extent from the redirector */
+                    buf_RDRShakeAnExtentFree(bp, reqp);
+                }
+                lock_ReleaseRead(&buf_globalLock);
             }
-            lock_ReleaseRead(&buf_globalLock);
 
             lock_ObtainMutex(&bp->mx);
             if ((bp->flags & CM_BUF_DIRTY)) {
@@ -2495,7 +2541,7 @@ long buf_RDRBuffersExist(cm_fid_t *fidp)
     afs_uint32 i;
     long found = 0;
 
-    if (!RDR_Initialized)
+    if (!RDR_Initialized || cm_directIO)
         return 0;
 
     i = BUF_FILEHASH(fidp);
@@ -2518,6 +2564,9 @@ long buf_ClearRDRFlag(cm_scache_t *scp, char *reason)
     afs_uint32 bcount = 0;
     afs_uint32 i;
 
+    if (!RDR_Initialized || cm_directIO)
+        return 0;
+
     i = BUF_FILEHASH(fidp);
 
     lock_ObtainWrite(&scp->rw);
