@@ -227,6 +227,8 @@ cm_callbackDiscardROVolumeByFID(cm_fid_t *fidp)
         if (volp->cbExpiresRO) {
             volp->cbExpiresRO = 0;
             volp->cbIssuedRO = 0;
+            volp->volumeSizeRO = 0;
+            _InterlockedAnd(&volp->flags, ~CM_VOLUMEFLAG_RO_SIZE_VALID);
             if (volp->cbServerpRO) {
                 cm_PutServer(volp->cbServerpRO);
                 volp->cbServerpRO = NULL;
@@ -1824,7 +1826,7 @@ long cm_GetCallback(cm_scache_t *scp, struct cm_user *userp,
             InterlockedIncrement(&scp->activeRPCs);
 
             // Fetch the status info
-            cm_MergeStatus(NULL, scp, &afsStatus, &volSync, userp, reqp, 0);
+            code = cm_MergeStatus(NULL, scp, &afsStatus, &volSync, userp, reqp, 0);
         }
         goto done;
     }
@@ -1865,7 +1867,7 @@ long cm_GetCallback(cm_scache_t *scp, struct cm_user *userp,
                                      &afsStatus, &callback, &volSync);
             rx_PutConnection(rxconnp);
 
-        } while (cm_Analyze(connp, userp, reqp, &sfid, NULL, 0, &volSync, NULL,
+        } while (cm_Analyze(connp, userp, reqp, &sfid, NULL, 0, &afsStatus, &volSync, NULL,
                             &cbr, code));
         code = cm_MapRPCError(code, reqp);
         if (code)
@@ -1878,8 +1880,11 @@ long cm_GetCallback(cm_scache_t *scp, struct cm_user *userp,
         lock_ObtainWrite(&scp->rw);
         if (code == 0) {
             int lostRace = cm_EndCallbackGrantingCall(scp, &cbr, &callback, &volSync, 0);
-            if (!lostRace)
-                cm_MergeStatus(NULL, scp, &afsStatus, &volSync, userp, reqp, 0);
+	    if (lostRace) {
+		InterlockedDecrement(&scp->activeRPCs);
+	    } else {
+                code = cm_MergeStatus(NULL, scp, &afsStatus, &volSync, userp, reqp, 0);
+	    }
         } else {
             cm_EndCallbackGrantingCall(NULL, &cbr, NULL, NULL, 0);
             InterlockedDecrement(&scp->activeRPCs);
