@@ -866,10 +866,9 @@ cm_Analyze(cm_conn_t *connp,
 		    pscp = cm_FindSCacheParent(scp);
 
 		lock_ObtainWrite(&scp->rw);
-		scp->flags |= CM_SCACHEFLAG_DELETED;
+		_InterlockedOr(&scp->flags, CM_SCACHEFLAG_DELETED);
 		lock_ObtainWrite(&cm_scacheLock);
                 cm_AdjustScacheLRU(scp);
-                cm_RemoveSCacheFromHashTable(scp);
 		lock_ReleaseWrite(&cm_scacheLock);
                 cm_LockMarkSCacheLost(scp);
 		lock_ReleaseWrite(&scp->rw);
