@@ -547,7 +547,7 @@ DNew(struct dcache *adc, int page, struct DirBuffer *entry)
      * DFlush due to lock hierarchy issues */
     if ((page + 1) * AFS_BUFFER_PAGESIZE > adc->f.chunkBytes) {
 	afs_AdjustSize(adc, (page + 1) * AFS_BUFFER_PAGESIZE);
-	osi_Assert(afs_WriteDCache(adc, 1) == 0);
+	afs_WriteDCache(adc, 1);
     }
     ObtainWriteLock(&tb->lock, 265);
     tb->lockers++;
